@@ -5,7 +5,7 @@
  *	Copyright (C) 2002 ActiveState Corporation
  *	Copyright (C) 2004 Starfish Systems 
  *
- * $Header: /cvsroot/tls/tls/tls.c,v 1.30 2008/03/19 22:06:13 hobbs2 Exp $
+ * $Header: /cvsroot/tls/tls/tls.c,v 1.35 2014/12/08 19:09:06 andreas_kupries Exp $
  *
  * TLS (aka SSL) Channel - can be layered on any bi-directional
  * Tcl_Channel (Note: Requires Trf Core Patch)
@@ -66,9 +66,13 @@ static int	UnimportObjCmd _ANSI_ARGS_ ((ClientData clientData,
 static SSL_CTX *CTX_Init _ANSI_ARGS_((State *statePtr, int proto, char *key,
 			char *cert, char *CAdir, char *CAfile, char *ciphers));
 
-#define TLS_PROTO_SSL2	0x01
-#define TLS_PROTO_SSL3	0x02
-#define TLS_PROTO_TLS1	0x04
+static int	TlsLibInit _ANSI_ARGS_ (()) ;
+
+#define TLS_PROTO_SSL2		0x01
+#define TLS_PROTO_SSL3		0x02
+#define TLS_PROTO_TLS1		0x04
+#define TLS_PROTO_TLS1_1	0x08
+#define TLS_PROTO_TLS1_2	0x10
 #define ENABLED(flag, mask)	(((flag) & (mask)) == (mask))
 
 /*
@@ -130,6 +134,47 @@ int channelTypeVersion;
 #define sk_SSL_CIPHER_value( sk, index)	(SSL_CIPHER*)sk_value((sk), (index))
 #endif
 
+/*
+ * Thread-Safe TLS Code
+ */
+
+#ifdef TCL_THREADS
+#define OPENSSL_THREAD_DEFINES
+#include <openssl/opensslconf.h>
+
+#ifdef OPENSSL_THREADS
+#include <openssl/crypto.h>
+
+/*
+ * Threaded operation requires locking callbacks
+ * Based from /crypto/cryptlib.c of OpenSSL and NSOpenSSL.
+ */
+
+static Tcl_Mutex locks[CRYPTO_NUM_LOCKS];
+static Tcl_Mutex init_mx;
+static int initialized;
+
+static void          CryptoThreadLockCallback (int mode, int n, const char *file, int line);
+static unsigned long CryptoThreadIdCallback   (void);
+
+static void
+CryptoThreadLockCallback(int mode, int n, const char *file, int line)
+{
+    if (mode & CRYPTO_LOCK) {
+       Tcl_MutexLock(&locks[n]);
+    } else {
+       Tcl_MutexUnlock(&locks[n]);
+    }
+}
+
+static unsigned long
+CryptoThreadIdCallback(void)
+{
+    return (unsigned long) Tcl_GetCurrentThread();
+}
+#endif /* OPENSSL_THREADS */
+#endif /* TCL_THREADS */
+
 
 /*
  *-------------------------------------------------------------------
@@ -465,10 +510,10 @@ CiphersObjCmd(clientData, interp, objc, objv)
     Tcl_Obj	*CONST objv[];
 {
     static CONST84 char *protocols[] = {
-	"ssl2",	"ssl3",	"tls1",	NULL
+	"ssl2",	"ssl3",	"tls1",	"tls1.1", "tls1.2", NULL
     };
     enum protocol {
-	TLS_SSL2, TLS_SSL3, TLS_TLS1, TLS_NONE
+	TLS_SSL2, TLS_SSL3, TLS_TLS1, TLS_TLS1_1, TLS_TLS1_2, TLS_NONE
     };
     Tcl_Obj *objPtr;
     SSL_CTX *ctx = NULL;
@@ -510,6 +555,20 @@ CiphersObjCmd(clientData, interp, objc, objv)
 		return TCL_ERROR;
 #else
 		ctx = SSL_CTX_new(TLSv1_method()); break;
+#endif
+    case TLS_TLS1_1:
+#if defined(NO_TLS1_1)
+		Tcl_AppendResult(interp, "protocol not supported", NULL);
+		return TCL_ERROR;
+#else
+		ctx = SSL_CTX_new(TLSv1_1_method()); break;
+#endif
+    case TLS_TLS1_2:
+#if defined(NO_TLS1_2)
+		Tcl_AppendResult(interp, "protocol not supported", NULL);
+		return TCL_ERROR;
+#else
+		ctx = SSL_CTX_new(TLSv1_2_method()); break;
 #endif
     default:
 		break;
@@ -673,6 +732,9 @@ ImportObjCmd(clientData, interp, objc, objv)
     char *CAfile	= NULL;
     char *CAdir		= NULL;
     char *model		= NULL;
+#ifndef OPENSSL_NO_TLSEXT
+    char *servername	= NULL;	/* hostname for Server Name Indication */
+#endif
 #if defined(NO_SSL2)
     int ssl2 = 0;
 #else
@@ -683,11 +745,9 @@ ImportObjCmd(clientData, interp, objc, objv)
 #else
     int ssl3 = 1;
 #endif
-#if defined(NO_SSL2) && defined(NO_SSL3)
     int tls1 = 1;
-#else
-    int tls1 = 0;
-#endif
+    int tls1_1 = 1;
+    int tls1_2 = 1;
     int proto = 0;
     int verify = 0, require = 0, request = 1;
 
@@ -724,12 +784,17 @@ ImportObjCmd(clientData, interp, objc, objv)
 	OPTBOOL( "-require", require);
 	OPTBOOL( "-request", request);
 	OPTBOOL( "-server", server);
+#ifndef OPENSSL_NO_TLSEXT
+        OPTSTR( "-servername", servername);
+#endif
 
 	OPTBOOL( "-ssl2", ssl2);
 	OPTBOOL( "-ssl3", ssl3);
 	OPTBOOL( "-tls1", tls1);
+	OPTBOOL( "-tls1.1", tls1_1);
+	OPTBOOL( "-tls1.2", tls1_2);
 
-	OPTBAD( "option", "-cadir, -cafile, -certfile, -cipher, -command, -keyfile, -model, -password, -require, -request, -server, -ssl2, -ssl3, or -tls1");
+	OPTBAD( "option", "-cadir, -cafile, -certfile, -cipher, -command, -keyfile, -model, -password, -require, -request, -server, -servername, -ssl2, -ssl3, -tls1, -tls1.1 or -tls1.2");
 
 	return TCL_ERROR;
     }
@@ -740,6 +805,8 @@ ImportObjCmd(clientData, interp, objc, objv)
     proto |= (ssl2 ? TLS_PROTO_SSL2 : 0);
     proto |= (ssl3 ? TLS_PROTO_SSL3 : 0);
     proto |= (tls1 ? TLS_PROTO_TLS1 : 0);
+    proto |= (tls1_1 ? TLS_PROTO_TLS1_1 : 0);
+    proto |= (tls1_2 ? TLS_PROTO_TLS1_2 : 0);
 
     /* reset to NULL if blank string provided */
     if (cert && !*cert)		cert	= NULL;
@@ -846,6 +913,17 @@ ImportObjCmd(clientData, interp, objc, objv)
 	return TCL_ERROR;
     }
 
+#ifndef OPENSSL_NO_TLSEXT
+    if (servername) {
+        if (!SSL_set_tlsext_host_name(statePtr->ssl, servername) && require) {
+            Tcl_AppendResult(interp, "setting TLS host name extension failed",
+                (char *) NULL);
+            Tls_Free((char *) statePtr);
+            return TCL_ERROR;
+        }
+    }
+#endif
+
     /*
      * SSL Callbacks
      */
@@ -961,38 +1039,93 @@ CTX_Init(statePtr, proto, key, cert, CAdir, CAfile, ciphers)
     Tcl_DString ds;
     Tcl_DString ds1;
     int off = 0;
+    const SSL_METHOD *method;
 
-    /* create SSL context */
-#if !defined(NO_SSL2) && !defined(NO_SSL3)
-    if (ENABLED(proto, TLS_PROTO_SSL2) &&
-	ENABLED(proto, TLS_PROTO_SSL3)) {
-	ctx = SSL_CTX_new(SSLv23_method());
-    } else
-#endif
-    if (ENABLED(proto, TLS_PROTO_SSL2)) {
-#if defined(NO_SSL2)
-	Tcl_AppendResult(interp, "protocol not supported", NULL);
-	return (SSL_CTX *)0;
-#else
-	ctx = SSL_CTX_new(SSLv2_method());
-#endif
-    } else if (ENABLED(proto, TLS_PROTO_TLS1)) {
-	ctx = SSL_CTX_new(TLSv1_method());
-    } else if (ENABLED(proto, TLS_PROTO_SSL3)) {
-#if defined(NO_SSL3)
-	Tcl_AppendResult(interp, "protocol not supported", NULL);
-	return (SSL_CTX *)0;
-#else
-	ctx = SSL_CTX_new(SSLv3_method());
-#endif
-    } else {
+    if (!proto) {
 	Tcl_AppendResult(interp, "no valid protocol selected", NULL);
 	return (SSL_CTX *)0;
     }
-    off |= (ENABLED(proto, TLS_PROTO_TLS1) ? 0 : SSL_OP_NO_TLSv1);
-    off |= (ENABLED(proto, TLS_PROTO_SSL2) ? 0 : SSL_OP_NO_SSLv2);
-    off |= (ENABLED(proto, TLS_PROTO_SSL3) ? 0 : SSL_OP_NO_SSLv3);
 
+    /* create SSL context */
+#if defined(NO_SSL2)
+    if (ENABLED(proto, TLS_PROTO_SSL2)) {
+	Tcl_AppendResult(interp, "protocol not supported", NULL);
+	return (SSL_CTX *)0;
+    }
+#endif
+#if defined(NO_SSL3)
+    if (ENABLED(proto, TLS_PROTO_SSL3)) {
+	Tcl_AppendResult(interp, "protocol not supported", NULL);
+	return (SSL_CTX *)0;
+    }
+#endif
+#if defined(NO_TLS1)
+    if (ENABLED(proto, TLS_PROTO_TLS1)) {
+	Tcl_AppendResult(interp, "protocol not supported", NULL);
+	return (SSL_CTX *)0;
+    }
+#endif
+#if defined(NO_TLS1_1)
+    if (ENABLED(proto, TLS_PROTO_TLS1_1)) {
+	Tcl_AppendResult(interp, "protocol not supported", NULL);
+	return (SSL_CTX *)0;
+    }
+#endif
+#if defined(NO_TLS1_2)
+    if (ENABLED(proto, TLS_PROTO_TLS1_2)) {
+	Tcl_AppendResult(interp, "protocol not supported", NULL);
+	return (SSL_CTX *)0;
+    }
+#endif
+
+    switch (proto) {
+#if !defined(NO_SSL2)
+    case TLS_PROTO_SSL2:
+	method = SSLv2_method ();
+	break;
+#endif
+#if !defined(NO_SSL3)
+    case TLS_PROTO_SSL3:
+	method = SSLv3_method ();
+	break;
+#endif
+#if !defined(NO_TLS1)
+    case TLS_PROTO_TLS1:
+	method = TLSv1_method ();
+	break;
+#endif
+#if !defined(NO_TLS1_1)
+    case TLS_PROTO_TLS1_1:
+	method = TLSv1_1_method ();
+	break;
+#endif
+#if !defined(NO_TLS1_2)
+    case TLS_PROTO_TLS1_2:
+	method = TLSv1_2_method ();
+	break;
+#endif
+    default:
+        method = SSLv23_method ();
+#if !defined(NO_SSL2)
+	off |= (ENABLED(proto, TLS_PROTO_SSL2)   ? 0 : SSL_OP_NO_SSLv2);
+#endif
+#if !defined(NO_SSL3)
+	off |= (ENABLED(proto, TLS_PROTO_SSL3)   ? 0 : SSL_OP_NO_SSLv3);
+#endif
+#if !defined(NO_TLS1)
+	off |= (ENABLED(proto, TLS_PROTO_TLS1)   ? 0 : SSL_OP_NO_TLSv1);
+#endif
+#if !defined(NO_TLS1_1)
+	off |= (ENABLED(proto, TLS_PROTO_TLS1_1) ? 0 : SSL_OP_NO_TLSv1_1);
+#endif
+#if !defined(NO_TLS1_2)
+	off |= (ENABLED(proto, TLS_PROTO_TLS1_2) ? 0 : SSL_OP_NO_TLSv1_2);
+#endif
+	break;
+    }
+
+    ctx = SSL_CTX_new (method);
+    
     SSL_CTX_set_app_data( ctx, (VOID*)interp);	/* remember the interpreter */
     SSL_CTX_set_options( ctx, SSL_OP_ALL);	/* all SSL bug workarounds */
     SSL_CTX_set_options( ctx, off);	/* all SSL bug workarounds */
@@ -1149,13 +1282,15 @@ StatusObjCmd(clientData, interp, objc, objv)
 		"\": not a TLS channel", NULL);
 	return TCL_ERROR;
     }
-    statePtr	= (State *) Tcl_GetChannelInstanceData(chan);
-    if (objc == 2)
-	peer	= SSL_get_peer_certificate(statePtr->ssl);
-    else
-	peer	= SSL_get_certificate(statePtr->ssl);
+    statePtr = (State *) Tcl_GetChannelInstanceData(chan);
+    if (objc == 2) {
+	peer = SSL_get_peer_certificate(statePtr->ssl);
+    } else {
+	peer = SSL_get_certificate(statePtr->ssl);
+    }
     if (peer) {
 	objPtr = Tls_NewX509Obj(interp, peer);
+	if (objc == 2) { X509_free(peer); }
     } else {
 	objPtr = Tcl_NewListObj(0, NULL);
     }
@@ -1224,7 +1359,7 @@ MiscObjCmd(clientData, interp, objc, objv)
     int objc;
     Tcl_Obj	*CONST objv[];
 {
-    CONST84 char *commands [] = { "req", NULL };
+    static CONST84 char *commands [] = { "req", NULL };
     enum command { C_REQ, C_DUMMY };
     int cmd;
 
@@ -1466,8 +1601,7 @@ int
 Tls_Init(Tcl_Interp *interp)		/* Interpreter in which the package is
 					 * to be made available. */
 {
-    int major, minor, patchlevel, release, i;
-    char rnd_seed[16] = "GrzSlplKqUdnnzP!";	/* 16 bytes */
+    int major, minor, patchlevel, release;
 
     /*
      * The original 8.2.0 stacked channel implementation (and the patch
@@ -1500,30 +1634,10 @@ Tls_Init(Tcl_Interp *interp)		/* Interpreter in which the package is
 	channelTypeVersion = TLS_CHANNEL_VERSION_1;
     }
 
-    if (SSL_library_init() != 1) {
+    if (TlsLibInit() != TCL_OK) {
 	Tcl_AppendResult(interp, "could not initialize SSL library", NULL);
 	return TCL_ERROR;
     }
-    SSL_load_error_strings();
-    ERR_load_crypto_strings();
-
-    /*
-     * Seed the random number generator in the SSL library,
-     * using the do/while construct because of the bug note in the
-     * OpenSSL FAQ at http://www.openssl.org/support/faq.html#USER1
-     *
-     * The crux of the problem is that Solaris 7 does not have a 
-     * /dev/random or /dev/urandom device so it cannot gather enough
-     * entropy from the RAND_seed() when TLS initializes and refuses
-     * to go further. Earlier versions of OpenSSL carried on regardless.
-     */
-    srand((unsigned int) time((time_t *) NULL));
-    do {
-	for (i = 0; i < 16; i++) {
-	    rnd_seed[i] = 1 + (char) (255.0 * rand()/(RAND_MAX+1.0));
-	}
-	RAND_seed(rnd_seed, sizeof(rnd_seed));
-    } while (RAND_status() != 1);
 
     Tcl_CreateObjCommand(interp, "tls::ciphers", CiphersObjCmd,
 	    (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
@@ -1573,3 +1687,86 @@ Tls_SafeInit (Tcl_Interp* interp)
 {
     return Tls_Init (interp);
 }
+
+
+/*
+ *------------------------------------------------------*
+ *
+ *	TlsLibInit --
+ *
+ *	------------------------------------------------*
+ *	Initializes SSL library once per application
+ *	------------------------------------------------*
+ *
+ *	Side effects:
+ *		initilizes SSL library
+ *
+ *	Result:
+ *		none
+ *
+ *------------------------------------------------------*
+ */
+static int
+TlsLibInit ()
+{
+    int i;
+    char rnd_seed[16] = "GrzSlplKqUdnnzP!";	/* 16 bytes */
+#if defined(OPENSSL_THREADS) && defined(TCL_THREADS)
+    size_t num_locks;
+#endif
+    int status=TCL_OK;
+
+    if (!initialized) {
+	Tcl_MutexLock(&init_mx);
+	if (!initialized) {
+	    initialized = 1;
+
+	    if (CRYPTO_set_mem_functions((void *(*)(size_t))Tcl_Alloc,
+					 (void *(*)(void *, size_t))Tcl_Realloc,
+					 (void(*)(void *))Tcl_Free) == 0) {
+	       /* Not using Tcl's mem functions ... not critical */
+	    }
+
+#if defined(OPENSSL_THREADS) && defined(TCL_THREADS)
+	    /* should we consider allocating mutexes? */
+	    num_locks = CRYPTO_num_locks();
+	    if (num_locks > CRYPTO_NUM_LOCKS) {
+		status=TCL_ERROR;
+		goto done;
+	    }
+
+	    CRYPTO_set_locking_callback(CryptoThreadLockCallback);
+	    CRYPTO_set_id_callback(CryptoThreadIdCallback);
+#endif
+
+	    if (SSL_library_init() != 1) {
+	    	status=TCL_ERROR;
+		goto done;
+	    }
+	    SSL_load_error_strings();
+	    ERR_load_crypto_strings();
+
+	    /*
+	     * Seed the random number generator in the SSL library,
+	     * using the do/while construct because of the bug note in the
+	     * OpenSSL FAQ at http://www.openssl.org/support/faq.html#USER1
+	     *
+	     * The crux of the problem is that Solaris 7 does not have a 
+	     * /dev/random or /dev/urandom device so it cannot gather enough
+	     * entropy from the RAND_seed() when TLS initializes and refuses
+	     * to go further. Earlier versions of OpenSSL carried on regardless.
+	     */
+	    srand((unsigned int) time((time_t *) NULL));
+	    do {
+		for (i = 0; i < 16; i++) {
+		    rnd_seed[i] = 1 + (char) (255.0 * rand()/(RAND_MAX+1.0));
+		}
+		RAND_seed(rnd_seed, sizeof(rnd_seed));
+	    } while (RAND_status() != 1);
+	}
+    	done:
+
+	Tcl_MutexUnlock(&init_mx);
+    }
+    return status;
+}
